
/* Chrysalide - Outil d'analyse de fichiers binaires
 * counter.c - décompte de correspondances identifiées dans du contenu binaire
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "counter.h"


#include "counter-int.h"



/* --------------------- INSTANCIATION D'UNE FORME DE CONDITION --------------------- */


/* Initialise la classe des opérations booléennes. */
static void g_match_counter_class_init(GMatchCounterClass *);

/* Initialise une instance d'opération booléenne. */
static void g_match_counter_init(GMatchCounter *);

/* Supprime toutes les références externes. */
static void g_match_counter_dispose(GMatchCounter *);

/* Procède à la libération totale de la mémoire. */
static void g_match_counter_finalize(GMatchCounter *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Indique le statut d'une condition de validation. */
static unsigned long long g_match_counter_resolve_as_number(const GMatchCounter *);

/* Lance l'analyse de contenu binaire selon un motif donné. */
static void g_match_counter_analyze(const GMatchCounter *, const bin_t *, phys_t, phys_t, bool);



/* ---------------------------------------------------------------------------------- */
/*                       INSTANCIATION D'UNE FORME DE CONDITION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un décompte de résultats lors d'une recherche de motifs. */
G_DEFINE_TYPE(GMatchCounter, g_match_counter, G_TYPE_MATCH_CONDITION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des opérations booléennes.              *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_match_counter_class_init(GMatchCounterClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GMatchConditionClass *cond;             /* Classe parente directe      */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_match_counter_dispose;
    object->finalize = (GObjectFinalizeFunc)g_match_counter_finalize;

    cond = G_MATCH_CONDITION_CLASS(klass);

    cond->resolve_as_num = (resolve_cond_as_number_fc)g_match_counter_resolve_as_number;
    cond->analyze = (analyze_cond_fc)g_match_counter_analyze;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance à initialiser.                                 *
*                                                                             *
*  Description : Initialise une instance d'opération booléenne.               *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_match_counter_init(GMatchCounter *counter)
{
    counter->pattern = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_match_counter_dispose(GMatchCounter *counter)
{
    g_clear_object(&counter->pattern);

    G_OBJECT_CLASS(g_match_counter_parent_class)->dispose(G_OBJECT(counter));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_match_counter_finalize(GMatchCounter *counter)
{
    G_OBJECT_CLASS(g_match_counter_parent_class)->finalize(G_OBJECT(counter));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : pattern = motif à impliquer.                                 *
*                                                                             *
*  Description : Met en place un décompte de correspondances obtenues.        *
*                                                                             *
*  Retour      : Condition mise en place.                                     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GMatchCounter *g_match_counter_new(GSearchPattern *pattern)
{
    GMatchCounter *result;                  /* Structure à retourner       */

    result = g_object_new(G_TYPE_MATCH_COUNTER, NULL);

    result->pattern = pattern;
    g_object_ref(G_OBJECT(pattern));

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : op = condition à consulter.                                  *
*                                                                             *
*  Description : Indique le statut d'une condition de validation.             *
*                                                                             *
*  Retour      : Forme numérique de la condition considérée pour validation.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static unsigned long long g_match_counter_resolve_as_number(const GMatchCounter *counter)
{
    unsigned long long result;              /* Valeur à retourner          */

    result = g_search_pattern_count_matchs(counter->pattern);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : counter = condition à considérer.                            *
*                data    = données binaires brutes à considérer.              *
*                size    = quantité de ces données.                           *
*                pos     = position du point d'étude courant.                 *
*                full    = force une recherche pleine et entière.             *
*                                                                             *
*  Description : Lance l'analyse de contenu binaire selon un motif donné.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_match_counter_analyze(const GMatchCounter *counter, const bin_t *data, phys_t size, phys_t pos, bool full)
{
    //g_search_pattern_analyze(counter->pattern, data, size, pos);

}
