
/* Chrysalide - Outil d'analyse de fichiers binaires
 * arithmop.c - gestion des opérations arithmétiques
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "arithmop.h"


#include "arithmop-int.h"
#include "literal.h"



/* --------------------- INTRODUCTION D'UNE NOUVELLE EXPRESSION --------------------- */


/* Initialise la classe des opérations arithmétiques. */
static void g_arithmetic_operation_class_init(GArithmOperationClass *);

/* Initialise une instance d'opération arithmétique. */
static void g_arithmetic_operation_init(GArithmOperation *);

/* Supprime toutes les références externes. */
static void g_arithmetic_operation_dispose(GArithmOperation *);

/* Procède à la libération totale de la mémoire. */
static void g_arithmetic_operation_finalize(GArithmOperation *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Réalise une comparaison entre objets selon un critère précis. */
static bool g_arithmetic_operation_compare_rich(const GArithmOperation *, const GArithmOperation *, RichCmpOperation, bool *);

/* Initialise une instance d'opération de relation. */
static GScanExpression *g_arithmetic_operation_duplicate(const GArithmOperation *);

/* Réduit une expression à une forme plus simple. */
GScanExpression *g_arithmetic_operation_reduce(GArithmOperation *, GScanContext *, bool);



/* ---------------------------------------------------------------------------------- */
/*                       INTRODUCTION D'UNE NOUVELLE EXPRESSION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une opération de relation entre expressions. */
G_DEFINE_TYPE(GArithmOperation, g_arithmetic_operation, G_TYPE_SCAN_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des opérations arithmétiques.           *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_class_init(GArithmOperationClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanExpressionClass *expr;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_arithmetic_operation_dispose;
    object->finalize = (GObjectFinalizeFunc)g_arithmetic_operation_finalize;

    expr = G_SCAN_EXPRESSION_CLASS(klass);

    expr->cmp_rich = (compare_expr_rich_fc)g_arithmetic_operation_compare_rich;
    expr->dup = (dup_expr_fc)g_arithmetic_operation_duplicate;
    expr->reduce = (reduce_expr_fc)g_arithmetic_operation_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance à initialiser.                                 *
*                                                                             *
*  Description : Initialise une instance d'opération arithmétique.            *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_init(GArithmOperation *op)
{
    op->first = NULL;
    op->second = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_dispose(GArithmOperation *op)
{
    g_clear_object(&op->first);
    g_clear_object(&op->second);

    G_OBJECT_CLASS(g_arithmetic_operation_parent_class)->dispose(G_OBJECT(op));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op = instance d'objet GLib à traiter.                        *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_arithmetic_operation_finalize(GArithmOperation *op)
{
    G_OBJECT_CLASS(g_arithmetic_operation_parent_class)->finalize(G_OBJECT(op));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : operator = type d'opération arithmétique à représenter.      *
*                first    = premier opérande concerné.                        *
*                second   = éventuel second opérande impliqué ou NULL.        *
*                                                                             *
*  Description : Organise une opération arithmétique entre expressions.       *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_arithmetic_operation_new(ArithmeticExpressionOperator operator, GScanExpression *first, GScanExpression *second)
{
    GScanExpression *result;                /* Structure à retourner       */

    result = g_object_new(G_TYPE_ARITHMETIC_OPERATION, NULL);

    if (!g_arithmetic_operation_create(G_ARITHMETIC_OPERATION(result), operator, first, second))
        g_clear_object(&result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : op       = instance à initialiser pleinement.                *
*                operator = type d'opération booléenne à représenter.         *
*                first    = premier opérande concerné.                        *
*                second   = éventuel second opérande impliqué ou NULL.        *
*                                                                             *
*  Description : Met en place une opération arithmétique entre expressions.   *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_arithmetic_operation_create(GArithmOperation *op, ArithmeticExpressionOperator operator, GScanExpression *first, GScanExpression *second)
{
    bool result;                            /* Bilan à retourner           */
    ExprValueType vtype;                    /* Type de valeur portée       */

    result = false;

    vtype = g_scan_expression_get_value_type(first);

    if (vtype != EVT_INTEGER && vtype != EVT_PENDING)
        goto exit;

    vtype = g_scan_expression_get_value_type(second);

    if (vtype != EVT_INTEGER && vtype != EVT_PENDING)
        goto exit;

    if (!g_scan_expression_create(G_SCAN_EXPRESSION(op), EVT_INTEGER))
        goto exit;

    op->operator = operator;

    op->first = first;
    g_object_ref(G_OBJECT(op->first));

    op->second = second;
    g_object_ref(G_OBJECT(op->second));

    result = true;

 exit:

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : item   = premier objet à consulter pour une comparaison.     *
*                other  = second objet à consulter pour une comparaison.      *
*                op     = opération de comparaison à réaliser.                *
*                status = bilan des opérations de comparaison. [OUT]          *
*                                                                             *
*  Description : Réalise une comparaison entre objets selon un critère précis.*
*                                                                             *
*  Retour      : true si la comparaison a pu être effectuée, false sinon.     *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static bool g_arithmetic_operation_compare_rich(const GArithmOperation *item, const GArithmOperation *other, RichCmpOperation op, bool *status)
{
    bool result;                            /* Etat à retourner            */
    bool equal;                             /* Bilan intermédiaire         */

    result = true; // TODO : cmp parent()->type

    if (item->operator != other->operator)
    {
        result = compare_rich_integer_values(item->operator, other->operator, op);
        goto done;
    }

    equal = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item), G_COMPARABLE_ITEM(other), RCO_EQ, status);

    if (!equal)
    {
        result = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item->first),
                                                G_COMPARABLE_ITEM(other->first),
                                                op, status);
        goto done;
    }

    result = g_comparable_item_compare_rich(G_COMPARABLE_ITEM(item->second),
                                            G_COMPARABLE_ITEM(other->second),
                                            op, status);

 done:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = expression à copier.                                  *
*                                                                             *
*  Description : Reproduit une expression en place dans une nouvelle instance.*
*                                                                             *
*  Retour      : Nouvelle instance d'expression.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GScanExpression *g_arithmetic_operation_duplicate(const GArithmOperation *expr)
{
    GScanExpression *result;                /* Instance copiée à retourner */

    result = g_arithmetic_operation_new(expr->operator, expr->first, expr->second);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = expression à consulter.                              *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                final = impose une conversion finale de dernier tour.        *
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Réduction correspondante, expression déjà réduite, ou NULL.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_arithmetic_operation_reduce(GArithmOperation *expr, GScanContext *ctx, bool final)
{
    GScanExpression *result;                /* Instance à renvoyer         */
    GScanExpression *new;                   /* Nouvelle expression obtenue */
    unsigned long long val_1;               /* Première valeur à traiter   */
    unsigned long long val_2;               /* Second valeur à traiter     */
    bool valid;                             /* Validité de ce bilan obtenu */
    unsigned long long reduced;             /* Valeur réduite finale       */

    result = NULL;

    /* Réduction des éléments considérés */

    new = g_scan_expression_reduce(expr->first, ctx, final);

    if (new != NULL)
    {
        g_object_unref(G_OBJECT(expr->first));
        expr->first = new;
    }

    if (expr->second != NULL)
    {
        new = g_scan_expression_reduce(expr->second, ctx, final);

        if (new != NULL)
        {
            g_object_unref(G_OBJECT(expr->second));
            expr->second = new;
        }

    }

    /* Construction d'une réduction locale ? */

    if (G_IS_LITERAL_EXPRESSION(expr->first) && G_IS_LITERAL_EXPRESSION(expr->second))
    {
        valid = g_literal_expression_get_integer_value(G_LITERAL_EXPRESSION(expr->first), &val_1);

        if (valid)
            valid = g_literal_expression_get_integer_value(G_LITERAL_EXPRESSION(expr->second), &val_2);

        if (valid)
            switch (expr->operator)
            {
                case AEO_PLUS:
                    reduced = val_1 + val_2;
                    break;

                case AEO_MINUS:
                    reduced = val_1 - val_2;
                    break;

                case AEO_MUL:
                    reduced = val_1 * val_2;
                    break;

                case AEO_DIV:
                    valid = (val_2 != 0);
                    if (valid)
                        reduced = val_1 / val_2;
                    break;

                case AEO_MOD:
                    valid = (val_2 != 0);
                    if (valid)
                        reduced = val_1 % val_2;
                    break;

            }

        if (valid)
            result = g_literal_expression_new(EVT_INTEGER, &reduced);

    }

    return result;

}
