
/* Chrysalide - Outil d'analyse de fichiers binaires
 * call.c - organisation d'un appel à un élément de scan enregistré
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "call.h"


#include <malloc.h>
#include <string.h>


#include "call-int.h"
#include "../../../core/global.h"



/* --------------------- INTRODUCTION D'UNE NOUVELLE EXPRESSION --------------------- */


/* Initialise la classe des appels de fonction avec arguments. */
static void g_pending_call_class_init(GPendingCallClass *);

/* Initialise une instance d'appel de fonction avec arguments. */
static void g_pending_call_init(GPendingCall *);

/* Supprime toutes les références externes. */
static void g_pending_call_dispose(GPendingCall *);

/* Procède à la libération totale de la mémoire. */
static void g_pending_call_finalize(GPendingCall *);

/* Définit une base de recherche pour la cible d'appel. */
static void g_pending_call_set_base(GPendingCall *, GRegisteredItem *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Reproduit une expression en place dans une nouvelle instance. */
static GScanExpression *g_pending_call_duplicate(const GPendingCall *);

/* Réduit une expression à une forme plus simple. */
GScanExpression *g_pending_call_reduce(GPendingCall *, GScanContext *, bool);



/* ---------------------------------------------------------------------------------- */
/*                       INTRODUCTION D'UNE NOUVELLE EXPRESSION                       */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour un appel de fonction enregistrée. */
G_DEFINE_TYPE(GPendingCall, g_pending_call, G_TYPE_SCAN_EXPRESSION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des appels de fonction avec arguments.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_class_init(GPendingCallClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GScanExpressionClass *expr;             /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_pending_call_dispose;
    object->finalize = (GObjectFinalizeFunc)g_pending_call_finalize;

    expr = G_SCAN_EXPRESSION_CLASS(klass);

    expr->cmp_rich = (compare_expr_rich_fc)NULL;
    expr->dup = (dup_expr_fc)g_pending_call_duplicate;
    expr->reduce = (reduce_expr_fc)g_pending_call_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance d'appel de fonction avec arguments.  *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_init(GPendingCall *call)
{
    call->base = NULL;
    call->target = NULL;

    call->args = NULL;
    call->count = 0;

    call->next = NULL;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_dispose(GPendingCall *call)
{
    size_t i;                               /* Boucle de parcours          */

    g_clear_object(&call->base);

    for (i = 0; i < call->count; i++)
        g_clear_object(&call->args[i]);

    g_clear_object(&call->next);

    G_OBJECT_CLASS(g_pending_call_parent_class)->dispose(G_OBJECT(call));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_finalize(GPendingCall *call)
{
    if (call->target != NULL)
        free(call->target);

    if (call->args != NULL)
        free(call->args);

    G_OBJECT_CLASS(g_pending_call_parent_class)->finalize(G_OBJECT(call));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : target = désignation de l'objet d'appel à identifier.        *
*                len    = taille de cette désignation.                        *
*                args   = éventuelle liste d'arguments à actionner.           *
*                count  = quantité de ces arguments.                          *
*                                                                             *
*  Description : Organise un appel de fonction avec ses arguments.            *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_pending_call_new(const char *target, size_t len, GScanExpression **args, size_t count)
{
    GScanExpression *result;                /* Structure à retourner       */

    result = g_object_new(G_TYPE_PENDING_CALL, NULL);

    if (!g_pending_call_create(G_PENDING_CALL(result), target, len, args, count))
        g_clear_object(&result);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call   = instance à initialiser pleinement.                  *
*                target = désignation de l'objet d'appel à identifier.        *
*                len    = taille de cette désignation.                        *
*                args   = éventuelle liste d'arguments à actionner.           *
*                count  = quantité de ces arguments.                          *
*                                                                             *
*  Description : Met en place une expression d'appel.                         *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_pending_call_create(GPendingCall *call, const char *target, size_t len, GScanExpression **args, size_t count)
{
    bool result;                            /* Bilan à retourner           */
    size_t i;                               /* Boucle de parcours          */

    result = g_scan_expression_create(G_SCAN_EXPRESSION(call), EVT_PENDING);
    if (!result) goto exit;

    call->target = strndup(target, len);

    call->args = malloc(count * sizeof(GScanExpression *));
    call->count = count;

    for (i = 0; i < count; i++)
    {
        call->args[i] = args[i];
        g_object_ref(G_OBJECT(args[i]));
    }

 exit:

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = expression d'appel à actualiser.                      *
*                base   = zone de recherche pour la résolution à venir.       *
*                                                                             *
*  Description : Définit une base de recherche pour la cible d'appel.         *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_pending_call_set_base(GPendingCall *call, GRegisteredItem *base)
{
    call->base = base;
    g_object_ref(G_OBJECT(base));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : call = expression d'appel à compléter.                       *
*                next = expression d'appel suivante dans la chaîne.           *
*                                                                             *
*  Description : Complète la chaine d'expressions d'appel.                    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

void g_pending_call_attach_next(GPendingCall *call, GPendingCall *next)
{
    call->next = next;
    g_object_ref(G_OBJECT(next));

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : expr = expression à copier.                                  *
*                                                                             *
*  Description : Reproduit une expression en place dans une nouvelle instance.*
*                                                                             *
*  Retour      : Nouvelle instance d'expression.                              *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GScanExpression *g_pending_call_duplicate(const GPendingCall *expr)
{
    GScanExpression *result;                /* Instance copiée à retourner */

    result = g_pending_call_new(expr->target, strlen(expr->target), expr->args, expr->count);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : expr  = expression à consulter.                              *
*                ctx   = contexte de suivi de l'analyse courante.             *
*                final = indique une ultime conversion dans le cycle en cours.*
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Réduction correspondante, expression déjà réduite, ou NULL.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_pending_call_reduce(GPendingCall *expr, GScanContext *ctx, bool final)
{
    GScanExpression *result;                /* Instance à renvoyer         */
    size_t i;                               /* Boucle de parcours          */
    GScanExpression *new;                   /* Nouvelle expression obtenue */
    bool last;                              /* Détection de fin de chaîne  */
    GRegisteredItem *base;                  /* Base de recherche courante  */
    GRegisteredItem *rebase;                /* Nouvelle base résolue       */
    GScanExpression *new_next;              /* Nouvelle version du suivant */

    for (i = 0; i < expr->count; i++)
    {
        new = g_scan_expression_reduce(expr->args[i], ctx, final);

        if (new != NULL)
        {
            g_object_unref(G_OBJECT(expr->args[i]));
            expr->args[i] = new;
        }

    }

    last = (expr->next == NULL);

    if (!last)
        new_next = g_scan_expression_duplicate(G_SCAN_EXPRESSION(expr->next));
    else
        new_next = NULL;

    if (expr->target != NULL)
    {
        if (expr->base != NULL)
        {
            base = expr->base;
            g_object_ref(G_OBJECT(base));
        }
        else
            base = G_REGISTERED_ITEM(get_rost_root_namespace());

        rebase = g_registered_item_resolve(base, expr->target, ctx,
                                           last ? NULL : expr->args, last ? 0 : expr->count,
                                           last, final);

        g_object_unref(G_OBJECT(base));

        if (rebase == NULL)
        {
            result = NULL;
            goto done;
        }

        if (last)
        {
            g_pending_call_set_base(expr, rebase);

            free(expr->target);
            expr->target = NULL;

        }
        else
            g_pending_call_set_base(G_PENDING_CALL(new_next), rebase);

    }

    if (last)
        result = g_registered_item_reduce(expr->base, ctx, expr->args, expr->count, final);

    else
    {
        result = g_scan_expression_reduce(new_next, ctx, final);

        if (result == NULL)
        {
            g_object_ref(G_OBJECT(new_next));
            result = new_next;
        }

    }

 done:

    g_clear_object(&new_next);

    return result;

}
