
/* Chrysalide - Outil d'analyse de fichiers binaires
 * datasize.c - récupération de la taille du contenu scanné
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "datasize.h"


#include "../func-int.h"
#include "../exprs/literal.h"



/* ---------------------- INTRODUCTION D'UNE NOUVELLE FONCTION ---------------------- */


/* Initialise la classe des mesures de quantité de données. */
static void g_datasize_function_class_init(GDatasizeFunctionClass *);

/* Initialise une instance de mesure de quantité de données. */
static void g_datasize_function_init(GDatasizeFunction *);

/* Supprime toutes les références externes. */
static void g_datasize_function_dispose(GDatasizeFunction *);

/* Procède à la libération totale de la mémoire. */
static void g_datasize_function_finalize(GDatasizeFunction *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Réduit une expression à une forme plus simple. */
static GScanExpression *g_datasize_function_reduce(GDatasizeFunction *, GScanContext *, GScanExpression **, size_t, bool);



/* ---------------------------------------------------------------------------------- */
/*                        INTRODUCTION D'UNE NOUVELLE FONCTION                        */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une mesure de quantité de données scannées. */
G_DEFINE_TYPE(GDatasizeFunction, g_datasize_function, G_TYPE_SCAN_FUNCTION);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des mesures de quantité de données.     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_datasize_function_class_init(GDatasizeFunctionClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GRegisteredItemClass *registered;       /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_datasize_function_dispose;
    object->finalize = (GObjectFinalizeFunc)g_datasize_function_finalize;

    registered = G_REGISTERED_ITEM_CLASS(klass);

    registered->reduce = (reduce_registered_item_fc)g_datasize_function_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance à initialiser.                               *
*                                                                             *
*  Description : Initialise une instance de mesure de quantité de données.    *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_datasize_function_init(GDatasizeFunction *func)
{

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_datasize_function_dispose(GDatasizeFunction *func)
{
    G_OBJECT_CLASS(g_datasize_function_parent_class)->dispose(G_OBJECT(func));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : func = instance d'objet GLib à traiter.                      *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_datasize_function_finalize(GDatasizeFunction *func)
{
    G_OBJECT_CLASS(g_datasize_function_parent_class)->finalize(G_OBJECT(func));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Constitue une fonction de récupération de taille de données. *
*                                                                             *
*  Retour      : Fonction mise en place.                                      *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanFunction *g_datasize_function_new(void)
{
    GScanFunction *result;                  /* Structure à retourner       */

    result = g_object_new(G_TYPE_DATASIZE_FUNCTION, NULL);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : func   = élément d'appel à consulter.                        *
*                target = désignation de l'objet d'appel à identifier.        *
*                ctx    = contexte de suivi de l'analyse courante.            *
*                args   = liste d'éventuels arguments fournis.                *
*                count  = taille de cette liste.                              *
*                last   = l'élément est-il le dernier d'une chaîne d'appels ? *
*                final = indique une ultime conversion dans le cycle en cours.*
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Réduction correspondante, expression déjà réduite, ou NULL.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static GScanExpression *g_datasize_function_reduce(GDatasizeFunction *func, GScanContext *ctx, GScanExpression **args, size_t count, bool final)
{
    GScanExpression *result;                /* Instance à renvoyer         */
    GBinContent *content;                   /* Contenu à manipuler         */
    phys_t size;                            /* Quantité de données liées   */

    content = g_scan_context_get_content(ctx);

    size = g_binary_content_compute_size(content);

    result = g_literal_expression_new(EVT_INTEGER, (unsigned long long []){ size });

    g_object_unref(G_OBJECT(content));

    return result;

}
