
/* Chrysalide - Outil d'analyse de fichiers binaires
 * space.c - définition d'un espace de noms pour les fonctions de scan
 *
 * Copyright (C) 2022 Cyrille Bagard
 *
 *  This file is part of Chrysalide.
 *
 *  Chrysalide is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Chrysalide is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Foobar.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "space.h"


#include <assert.h>
#include <string.h>


#include "space-int.h"



/* ------------------------- SOCLE POUR LES ESPACES DE NOMS ------------------------- */


/* Initialise la classe des espaces de noms pour scan. */
static void g_scan_namespace_class_init(GScanNamespaceClass *);

/* Initialise une instance d'espace de noms pour scan. */
static void g_scan_namespace_init(GScanNamespace *);

/* Supprime toutes les références externes. */
static void g_scan_namespace_dispose(GScanNamespace *);

/* Procède à la libération totale de la mémoire. */
static void g_scan_namespace_finalize(GScanNamespace *);



/* --------------------- IMPLEMENTATION DES FONCTIONS DE CLASSE --------------------- */


/* Lance une résolution d'élément à appeler.                    *
*                                                                             *
*  Retour      : Nouvel élément d'appel identifié ou NULL.                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GRegisteredItem *g_scan_namespace_resolve(GScanNamespace *, const char *, GScanContext *, GScanExpression **, size_t, bool, bool);

/* Réduit une expression à une forme plus simple. */
GScanExpression *g_scan_namespace_reduce(GScanNamespace *, GScanContext *, GScanExpression **, size_t, bool);



/* ---------------------------------------------------------------------------------- */
/*                           SOCLE POUR LES ESPACES DE NOMS                           */
/* ---------------------------------------------------------------------------------- */


/* Indique le type défini pour une définition d'espace de noms. */
G_DEFINE_TYPE(GScanNamespace, g_scan_namespace, G_TYPE_REGISTERED_ITEM);


/******************************************************************************
*                                                                             *
*  Paramètres  : klass = classe à initialiser.                                *
*                                                                             *
*  Description : Initialise la classe des espaces de noms pour scan.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_namespace_class_init(GScanNamespaceClass *klass)
{
    GObjectClass *object;                   /* Autre version de la classe  */
    GRegisteredItemClass *registered;       /* Version de classe parente   */

    object = G_OBJECT_CLASS(klass);

    object->dispose = (GObjectFinalizeFunc/* ! */)g_scan_namespace_dispose;
    object->finalize = (GObjectFinalizeFunc)g_scan_namespace_finalize;

    registered = G_REGISTERED_ITEM_CLASS(klass);

    registered->resolve = (resolve_registered_item_fc)g_scan_namespace_resolve;
    registered->reduce = (reduce_registered_item_fc)g_scan_namespace_reduce;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : space = instance à initialiser.                              *
*                                                                             *
*  Description : Initialise une instance d'espace de noms pour scan.          *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_namespace_init(GScanNamespace *space)
{
    space->children = NULL;
    space->names = NULL;
    space->count = 0;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : space = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Supprime toutes les références externes.                     *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_namespace_dispose(GScanNamespace *space)
{
    size_t i;                               /* Boucle de parcours          */

    for (i = 0; i < space->count; i++)
        g_clear_object(&space->children[i]);

    G_OBJECT_CLASS(g_scan_namespace_parent_class)->dispose(G_OBJECT(space));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : space = instance d'objet GLib à traiter.                     *
*                                                                             *
*  Description : Procède à la libération totale de la mémoire.                *
*                                                                             *
*  Retour      : -                                                            *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

static void g_scan_namespace_finalize(GScanNamespace *space)
{
    size_t i;                               /* Boucle de parcours          */

    if (space->children != NULL)
        free(space->children);

    for (i = 0; i < space->count; i++)
        free(space->names[i]);

    if (space->names != NULL)
        free(space->names);

    G_OBJECT_CLASS(g_scan_namespace_parent_class)->finalize(G_OBJECT(space));

}


/******************************************************************************
*                                                                             *
*  Paramètres  : -                                                            *
*                                                                             *
*  Description : Construit un nouvel espace de noms pour scan.                *
*                                                                             *
*  Retour      : Adresse de la structure mise en place.                       *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanNamespace *g_scan_namespace_new(void)
{
    GScanNamespace *result;                 /* Instance à retourner        */

    result = g_object_new(G_TYPE_SCAN_NAMESPACE, NULL);

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : space = espace de noms à compléter.                          *
*                type  = type d'élément à intégrer.                           *
*                name  = nom du futur espace ou NULL pour une racine.         *
*                                                                             *
*  Description : Intègre un nouvel élément dans l'esapce de noms.             *
*                                                                             *
*  Retour      : Bilan de l'opération.                                        *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

bool g_scan_namespace_register(GScanNamespace *space, GRegisteredItem *child, const char *name)
{
    bool result;                            /* Bilan à retourner           */

    result = true;

    space->count++;

    space->children = realloc(space->children, space->count * sizeof(GRegisteredItem *));
    space->children[space->count - 1] = child;

    space->names = realloc(space->names, space->count * sizeof(char *));
    space->names[space->count - 1] = strdup(name);

    return result;

}



/* ---------------------------------------------------------------------------------- */
/*                       IMPLEMENTATION DES FONCTIONS DE CLASSE                       */
/* ---------------------------------------------------------------------------------- */


/******************************************************************************
*                                                                             *
*  Paramètres  : space  = élément d'appel à consulter.                        *
*                target = désignation de l'objet d'appel à identifier.        *
*                ctx    = contexte de suivi de l'analyse courante.            *
*                args   = liste d'éventuels arguments fournis.                *
*                count  = taille de cette liste.                              *
*                last   = l'élément est-il le dernier d'une chaîne d'appels ? *
*                final = indique une ultime conversion dans le cycle en cours.*
*                                                                             *
*  Description : Lance une résolution d'élément à appeler.                    *
*                                                                             *
*  Retour      : Nouvel élément d'appel identifié ou NULL.                    *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GRegisteredItem *g_scan_namespace_resolve(GScanNamespace *space, const char *target, GScanContext *ctx, GScanExpression **args, size_t count, bool last, bool final)
{
    GRegisteredItem *result;                /* Instance à renvoyer         */
    size_t i;                               /* Boucle de parcours          */

    result = NULL;

    for (i = 0; i < space->count; i++)
        if (strcmp(target, space->names[i]) == 0)
        {
            result = space->children[i];
            g_object_ref(G_OBJECT(result));
            break;
        }

    return result;

}


/******************************************************************************
*                                                                             *
*  Paramètres  : space  = élément d'appel à consulter.                        *
*                ctx    = contexte de suivi de l'analyse courante.            *
*                args   = liste d'éventuels arguments fournis.                *
*                count  = taille de cette liste.                              *
*                final = indique une ultime conversion dans le cycle en cours.*
*                                                                             *
*  Description : Réduit une expression à une forme plus simple.               *
*                                                                             *
*  Retour      : Réduction correspondante, expression déjà réduite, ou NULL.  *
*                                                                             *
*  Remarques   : -                                                            *
*                                                                             *
******************************************************************************/

GScanExpression *g_scan_namespace_reduce(GScanNamespace *space, GScanContext *ctx, GScanExpression **args, size_t count, bool final)
{
    GScanExpression *result;                /* Instance à renvoyer         */

    assert(false);

    result = NULL;

    return result;

}
